Package: cito
Type: Package
Title: Building and Training Neural Networks
Version: 1.1
Authors@R: 
    c(person(given = "Christian",
           family = "Amesöder",
           role = c("aut"),
           email = "Christian.Amesoeder@informatik.uni-regensburg.de"),
      person(given = "Maximilian",
           family = "Pichler",
           role = c("aut", "cre"),
           email = "maximilian.pichler@biologie.uni-regensburg.de",
           comment = c(ORCID = "0000-0003-2252-8327")),
      person(given = "Florian",
             family = "Hartig",
             role = "ctb",
             email = "florian.hartig@biologie.uni-regensburg.de",
             comment = c(ORCID = "0000-0002-6255-9059")),           
      person(given = "Armin",
             family = "Schenk",
             role = "ctb",
             email = "armin.schenk99@gmail.com")
           )
Description: The 'cito' package provides a user-friendly interface for training and interpreting deep neural networks (DNN). 'cito' simplifies the fitting of DNNs by supporting the familiar formula syntax, hyperparameter tuning under cross-validation, and helps to detect and handle convergence problems.  DNNs can be trained on CPU, GPU and MacOS GPUs. In addition, 'cito' has many downstream functionalities such as various explainable AI (xAI) metrics (e.g. variable importance, partial dependence plots, accumulated local effect plots, and effect estimates) to interpret trained DNNs. 'cito' optionally provides confidence intervals (and p-values) for all xAI metrics and predictions. At the same time, 'cito' is computationally efficient because it is based on the deep learning framework 'torch'. The 'torch' package is native to R, so no Python installation or other API is required for this package.
Encoding: UTF-8
RoxygenNote: 7.2.3
Depends: R (>= 3.5)
Imports: coro, checkmate, torch, gridExtra, parabar, abind, progress,
        cli, torchvision, tibble, lme4
License: GPL (>= 3)
Suggests: spelling, rmarkdown, testthat, plotly, ggraph, igraph, stats,
        ggplot2, knitr
VignetteBuilder: knitr
BugReports: https://github.com/citoverse/cito/issues
URL: https://citoverse.github.io/cito/
Language: en-US
NeedsCompilation: no
Packaged: 2024-03-18 20:48:48 UTC; maximilianpichler
Author: Christian Amesöder [aut],
  Maximilian Pichler [aut, cre] (<https://orcid.org/0000-0003-2252-8327>),
  Florian Hartig [ctb] (<https://orcid.org/0000-0002-6255-9059>),
  Armin Schenk [ctb]
Maintainer: Maximilian Pichler <maximilian.pichler@biologie.uni-regensburg.de>
Repository: CRAN
Date/Publication: 2024-03-18 22:50:07 UTC
Built: R 4.6.0; ; 2026-04-29 00:41:52 UTC; unix
