# Generated by roxygen2: do not edit by hand

S3method(ALE,citodnn)
S3method(ALE,citodnnBootstrap)
S3method(PDP,citodnn)
S3method(PDP,citodnnBootstrap)
S3method(coef,citocnn)
S3method(coef,citodnn)
S3method(coef,citodnnBootstrap)
S3method(conditionalEffects,citodnn)
S3method(conditionalEffects,citodnnBootstrap)
S3method(continue_training,citocnn)
S3method(continue_training,citodnn)
S3method(continue_training,citodnnBootstrap)
S3method(plot,citocnn)
S3method(plot,citodnn)
S3method(plot,citodnnBootstrap)
S3method(predict,citocnn)
S3method(predict,citodnn)
S3method(predict,citodnnBootstrap)
S3method(print,citocnn)
S3method(print,citodnn)
S3method(print,citodnnBootstrap)
S3method(print,conditionalEffects)
S3method(print,conditionalEffectsBootstrap)
S3method(print,summary.citodnn)
S3method(print,summary.citodnnBootstrap)
S3method(residuals,citodnn)
S3method(summary,citocnn)
S3method(summary,citodnn)
S3method(summary,citodnnBootstrap)
export(ALE)
export(PDP)
export(analyze_training)
export(conditionalEffects)
export(config_lr_scheduler)
export(config_optimizer)
export(config_tuning)
export(continue_training)
export(dnn)
export(e)
export(tune)
import(checkmate)
import(tibble)
